import threading
from autoinject import injector
import datetime
import typing as t
from urllib.parse import urlparse


BYTE_UNITS = {
    # Bits
    "bit": 0.125,
    # Bytes
    "b": 1,
    # Standard prefixes
    "kib": 1024,
    "mib": 1048576,
    "gib": 1073741824,
    "tib": 1099511627776,
    "pib": 1125899906842624,
    "eib": 1152921504606846976,
    # Short forms translate to standard prefixes
    "k": 1024,
    "m": 1048576,
    "g": 1073741824,
    "t": 1099511627776,
    "p": 1125899906842624,
    "e": 1152921504606846976,
    # Allow for hard drive/metric sizing (but with option to disable)
    "kb": 1000,
    "mb": 1000000,
    "gb": 1000000000,
    "tb": 1000000000000,
    "pb": 1000000000000000,
    "eb": 1000000000000000000,
}


def parse_for_units(val: str, max_unit_len: int, default_units: str) -> t.Tuple[t.Union[int, float], str]:
    val = val.strip()
    if max_unit_len < 0:
        raise ValueError("Unit length must be positive")
    unit_len = -1 * min(max_unit_len, len(val))
    while unit_len < 0 and (val[unit_len].isdigit() or val[unit_len] == "."):
        unit_len += 1
    actual_val = val[:unit_len].strip() if unit_len < 0 else val
    units = val[unit_len:].strip() if unit_len < 0 else default_units
    return (float(actual_val), units) if "." in actual_val else (int(actual_val), units)


def convert_to_bytes(val: t.Union[float, int], units: str, disallow_metric_prefixes: bool = False) -> t.Union[float, int]:
    units = units.lower()
    # Convert metric prefixes to standard representations
    if disallow_metric_prefixes and len(units) == 2:
        units = f"{units[0]}i{units[1]}"
    if units not in BYTE_UNITS:
        raise ValueError(f"Unrecognized units for bytes: {units}")
    return val * BYTE_UNITS[units]


def convert_to_timedelta(val: t.Union[float, int], units: str) -> datetime.timedelta:
    units = units.lower()
    if units == "s":
        return datetime.timedelta(seconds=val)
    elif units == "m":
        return datetime.timedelta(minutes=val)
    elif units == "h":
        return datetime.timedelta(hours=val)
    elif units == "d":
        return datetime.timedelta(days=val)
    elif units == "w":
        return datetime.timedelta(weeks=val)
    elif units == "us":
        return datetime.timedelta(microseconds=val)
    elif units == "ms":
        return datetime.timedelta(milliseconds=val)
    else:
        raise ValueError(f"Unknown units for timedelta: {units}")


@injector.inject
def print_config(obfuscate_keys=None, cfg: "zirconium.config.ApplicationConfig" = None):
    print("----- Loaded Files -----")
    for f in cfg.loaded_files:
        print(f)
    print("----- Configuration Values -----")
    _print_dict(cfg.d, obfuscate_keys=obfuscate_keys)


def _print_dict(d, prefix='  ', level=0, parent_prefix=None, obfuscate_keys=None):
    for key in d:
        full_path = list(parent_prefix) if parent_prefix else []
        full_path.append(key)
        full_path = tuple(full_path)
        val = d[key]
        if obfuscate_keys is not None and full_path in obfuscate_keys:
            print(f"{prefix * level}{key}: {'*' * len(val)}")
            continue
        if isinstance(val, dict):
            print(f"{prefix * level}{key}: ")
            _print_dict(val, prefix, level + 1, full_path, obfuscate_keys)
        elif isinstance(val, set) or isinstance(val, list) or isinstance(val, tuple):
            print(f"{prefix * level}{key}: ")
            for entry in val:
                print(f"{prefix * (level+1)}- {_obfuscate_entry(entry)}")
        else:
            print(f"{prefix * level}{key}: {_obfuscate_entry(val)}")


def _obfuscate_entry(entry):
    if isinstance(entry, str) and "://" in entry:
        try:
            uri = urlparse(entry)
            if uri.password:
                qs = f"?{uri.query}" if uri.query else ""
                fs = f"#{uri.fragment}" if uri.fragment else ""
                return f"{uri.scheme}://{uri.username}:{'*' * len(uri.password)}@{uri.hostname}{uri.path}{qs}{fs}"
        except ValueError:
            pass
    return entry


@injector.injectable_global
class _AppConfigHooks:
    """ Global storage of configuration hooks for ApplicationConfig prior to instantiation. """

    def __init__(self):
        self.hooks = []

    def add_hook(self, c):
        self.hooks.append(c)

    def execute_hooks(self, cfg):
        for hook in self.hooks:
            hook(cfg)


@injector.inject
def _config_decorator(func, ach: _AppConfigHooks = None):
    """ Decorate a function with this to add configuration files """
    ach.add_hook(func)
    return func


class MutableDeepDict:
    """ Deep dictionary class that supports tuple-like access to deep properties """

    def __init__(self, base_dict=None):
        """ Constructor """
        self.d = base_dict if base_dict else {}
        self.lock = threading.RLock()

    def _navigate_to_item(self, key, create=False):
        """ Navigate to an item in the tree structure specified by key

            :param key: The key to navigate to
            :type key: str or tuple
            :param create: If set to true, the entry will be created
            :type create: bool
            :returns: A tuple, with the first item being the dictionary structure and the second being the key in that
                structure that represents the tail element of key. Both will be None if a parent key does not exist
            :rtype: tuple(dict, str)
        """
        try:
            if isinstance(key, str):
                return self.d, key
            parent = self.d
            for k in key[:-1]:
                if k in parent and MutableDeepDict.is_dict_like(parent[k]):
                    parent = parent[k]
                elif create:
                    parent[k] = {}
                    parent = parent[k]
                else:
                    return None, None
            return parent, key[-1]
        except TypeError:
            return self.d, key

    def __setitem__(self, key, value):
        """ Thread-safe __setitem__ implementation """
        with self.lock:
            parent, k = self._navigate_to_item(key, True)
            parent[k] = value

    def __getitem__(self, key):
        """ __getitem__ implementation"""
        return self.get(key, raise_error=True)

    def __delitem__(self, key):
        """ Thread-safe __delitem__ implementation"""
        with self.lock:
            parent, k = self._navigate_to_item(key)
            if parent:
                del parent[k]

    def __contains__(self, key):
        """ __contains__ implementation """
        parent, k = self._navigate_to_item(key)
        return parent is not None and k in parent

    def __len__(self):
        """ __len__ implementation """
        return len(self.d)

    def __iter__(self):
        """ __iter__ implementation """
        return iter(self.d)

    def clear(self):
        """Clear the dictionary of all entries."""
        self.d = {}

    def deep_update(self, d):
        """ Similar to update(), but will merge dictionaries at depth. Thread-safe. """
        with self.lock:
            for key in d.keys():
                if key in self.d and MutableDeepDict.is_dict_like(d[key]) and MutableDeepDict.is_dict_like(self.d[key]):
                    mut = MutableDeepDict(self.d[key])
                    mut.deep_update(d[key])
                else:
                    self.d[key] = d[key]

    def update(self, d):
        """ Thread-safe implementation of dict.update() """
        with self.lock:
            self.d.update(d)

    def _expand_key(self, key):
        """ Given a key, expands it to an ordered list to be used with _navigate_to_item() or other methods that
        leverage it """
        expanded = []
        for k in key:
            if isinstance(k, str) or not hasattr(k, "__iter__"):
                expanded.append(k)
            else:
                expanded.extend(k)
        return expanded

    def get(self, *key, default=None, raise_error=False):
        """ Implementation of dict.get(). Added a raise_error parameter which causes ValueError to be raised if the
            key does not exist, otherwise the default is returned. """
        key = self._expand_key(key)
        parent, k = self._navigate_to_item(key)
        if ((parent is None) or (not k in parent)) and raise_error:
            raise ValueError("No such key: {}".format(".".join(key)))
        value = default
        if parent and k in parent:
            value = parent[k]
        return value

    def keys(self):
        """ Implementation of dict.keys() """
        return self.d.keys()

    def values(self):
        """ Implementation of dict.values() """
        return self.d.values()

    def pop(self, key, default):
        """ Thread-safe implementation of dict.pop() that works on deep arrays. """
        with self.lock:
            parent, k = self._navigate_to_item(key)
            if parent:
                return parent.pop(key, default)
            return default

    @staticmethod
    def is_dict_like(d):
        """ Checks if d is dict-like """
        if isinstance(d, dict):
            return True
        if isinstance(d, MutableDeepDict):
             return True
        if hasattr(d, "keys"):
            return True
        return False



import unittest
import datetime
import decimal
import os
from pathlib import Path

from autoinject import injector
import zirconium


class TestConfig(unittest.TestCase):

    @injector.test_case()
    @zirconium.test_with_config(("foo", "bar"), "hello world")
    @zirconium.test_with_config(("foo", "bar"), "hello world2")
    @zirconium.test_with_config(("foo", "bar2"), "zoink")
    @zirconium.test_with_config({"foo": {"bar3": "hello world3"}})
    def test_test_case(self):
        @injector.inject
        def _test_inject(cfg: zirconium.ApplicationConfig = None):
            self.assertEqual(cfg.get(("foo", "bar")), "hello world")
            self.assertEqual(cfg.get(("foo", "bar2")), "zoink")
            self.assertEqual(cfg.get(("foo", "bar3")), "hello world3")
        _test_inject()

    def test_one_file(self):
        path = Path(__file__).parent / "example_configs/basic.yaml"
        config = zirconium.ApplicationConfig(True)
        config.register_file(path)
        config.init()
        self.assertTrue("one" in config)
        self.assertEqual(config["one"], "a")
        self.assertEqual(config["two"], 2)
        self.assertEqual(config["three"], "three")
        self.assertEqual(config["four"], " four five ")
        self.assertEqual(config["five"], datetime.date(2020, 1, 1))
        self.assertEqual(config["six"], 6.66)
        self.assertIsInstance(config["seven"], dict)
        self.assertEqual(len(config["seven"]), 2)
        self.assertIsInstance(config["twelve"], list)
        self.assertEqual(config["seventeen"], datetime.datetime(2020, 1, 1, 1, 2, 3))

    def test_two_files(self):
        path = Path(__file__).parent / "example_configs/basic.yaml"
        path2 = Path(__file__).parent / "example_configs/override.toml"
        config = zirconium.ApplicationConfig(True)
        config.set_defaults({
            "base": "one"
        })
        config.register_file(path)
        config.register_file(path2)
        config.init()
        self.assertEqual(config["base"], "one")
        self.assertTrue("one" in config)
        self.assertEqual(config["one"], "b")
        self.assertEqual(config["two"], 3)
        self.assertEqual(config["three"], "three")
        self.assertEqual(config["four"], " four five ")
        self.assertEqual(config["five"], datetime.date(2020, 1, 1))
        self.assertEqual(config["six"], 6.66)
        self.assertIsInstance(config["seven"], dict)
        self.assertEqual(len(config["seven"]), 2)
        self.assertIsInstance(config["twelve"], list)
        self.assertEqual(config["seventeen"], datetime.datetime(2020, 1, 1, 1, 2, 3))

    def test_environment_replacement(self):
        config = zirconium.ApplicationConfig(True)
        os.environ["VAR_NAME"] = "var"
        os.environ["lower_var"] = "var2"
        os.environ["${INNER}"] = "inner"
        os.environ["}end"] = "end"
        config.load_from_dict({
            "simple_replace": "${VAR_NAME}",
            "no_replace": "$${VAR_NAME}",
            "complex_replace": "$$$$${VAR_NAME}",
            "complex_noreplace": "$$$${VAR_NAME}",
            "default_use": "${VAR_NAME_2=test}",
            "no_end": "${VAR_NAME_NO_END",
            "suffix_replace": "${VAR_NAME} VAR_NAME",
            "prefix_replace": "VAR_NAME  ${VAR_NAME}",
            "middle_replace": "VN ${VAR_NAME} VN",
            "lower_var": "${LOWER_VAR}",
            "upper_var": "${var_name}",
            "complex_inner": r"${${INNER\}}",
            "weird_bracket": r"${\}end}",
            "weird_bracket_default": r"${\}end2=bar}"
        })

        self.assertEqual(config["simple_replace"], "var")
        self.assertEqual(config["default_use"], "test")
        self.assertEqual(config["complex_replace"], "$$var")
        self.assertEqual(config["complex_noreplace"], "$${VAR_NAME}")
        self.assertEqual(config["no_replace"], "${VAR_NAME}")
        self.assertEqual(config["no_end"], "${VAR_NAME_NO_END")
        self.assertEqual(config["lower_var"], "var2")
        self.assertEqual(config["upper_var"], "var")
        self.assertEqual(config["suffix_replace"], "var VAR_NAME")
        self.assertEqual(config["middle_replace"], "VN var VN")
        self.assertEqual(config["prefix_replace"], "VAR_NAME  var")
        self.assertEqual(config["weird_bracket"], "end")
        self.assertEqual(config["complex_inner"], "inner")
        self.assertEqual(config["weird_bracket_default"], "bar")

    def test_list_access(self):
        path = Path(__file__).parent / "example_configs/basic.yaml"
        config = zirconium.ApplicationConfig(True)
        config.register_file(path)
        config.init()
        self.assertEqual(config[("one",)], "a")
        self.assertEqual(config[("seven", "eight", "nine")], "nine")
        self.assertTrue(("seven", "eight") in config)
        self.assertFalse(("seven", "nine") in config)
        self.assertFalse(("test", "two") in config)
        config[("test", "two")] = "test three"
        self.assertTrue(("test", "two") in config)
        self.assertEqual(config[("test", "two")], "test three")
        self.assertEqual(config["test", "two"], "test three")
        self.assertEqual(config.get("test", "two"), "test three")
        del config["test", "two"]
        self.assertFalse(("test", "two") in config)

    def test_integer_index(self):
        config = zirconium.ApplicationConfig(True)
        config.load_from_dict({
            1: "one"
        })
        self.assertTrue(1 in config)
        self.assertFalse(2 in config)
        self.assertEqual(config[1], "one")

    def test_get_ref(self):
        config = zirconium.ApplicationConfig(True)
        config.set_defaults({
            1: "one"
        })
        config.init()
        r = config.get_ref(1)
        self.assertIsInstance(r, zirconium.config._ConfigRef)
        self.assertEqual(r.raw_value(), "one")
        self.assertEqual(r.raw_value(), "one")
        config.load_from_dict({
            1: "two"
        })
        self.assertEqual(r.raw_value(), "one")
        config.reload_config()
        config.load_from_dict({
            1: "two"
        })
        self.assertEqual(r.raw_value(), "two")

    def test_clear(self):
        config = zirconium.ApplicationConfig(True)
        config.load_from_dict({
            1: "one",
            2: "two"
        })
        self.assertTrue(1 in config)
        self.assertTrue(2 in config)
        self.assertEqual(config[1], "one")
        self.assertEqual(config[2], "two")
        config.reload_config()
        self.assertFalse(1 in config)
        self.assertFalse(2 in config)
        config.load_from_dict({
            1: "new_one"
        })
        self.assertTrue(1 in config)
        self.assertFalse(2 in config)
        self.assertEqual(config[1], "new_one")

    def test_len(self):
        config = zirconium.ApplicationConfig(True)
        self.assertEqual(len(config), 0)
        config.load_from_dict({
            "one": 1,
            "two": 2,
            "three": {
                "four": 4
            }
        })
        self.assertEqual(len(config), 3)

    def test_iter(self):
        config = zirconium.ApplicationConfig(True)
        config.load_from_dict({
            "one": 1,
            "two": 2,
            "three": {
                "four": 4
            }
        })
        lst = [x for x in config]
        self.assertTrue("one" in lst)
        self.assertTrue("two" in lst)
        self.assertTrue("three" in lst)
        self.assertFalse("four" in lst)

    def test_deep_update(self):
        config = zirconium.ApplicationConfig(True)
        config.load_from_dict({
            "one": 1,
            "two": 2,
            "three": {
                "four": 4
            }
        })
        self.assertTrue("one" in config)
        self.assertTrue("two" in config)
        self.assertTrue("three" in config)
        self.assertTrue(("three", "four") in config)
        self.assertFalse("four" in config)
        config.deep_update({
            "one": 11,
            "three": {
                "four": 44,
                "five": 55,
            },
            "six": 66
        })
        self.assertEqual(config["one"], 11)
        self.assertEqual(config["three"]["four"], 44)
        self.assertEqual(config["three"]["five"], 55)
        self.assertEqual(config["six"], 66)
        self.assertEqual(config["two"], 2)

    def test_update(self):
        config = zirconium.ApplicationConfig(True)
        config.load_from_dict({
            "one": 1,
            "two": 2,
            "three": {
                "four": 4,
                "seven": 7,
            }
        })
        config.update({
            "one": 11,
            "three": {
                "four": 44,
                "five": 55,
            },
            "six": 66
        })
        self.assertEqual(config["one"], 11)
        self.assertEqual(config["three"]["four"], 44)
        self.assertEqual(config["three"]["five"], 55)
        self.assertFalse("seven" in config["three"])
        self.assertEqual(config["six"], 66)
        self.assertEqual(config["two"], 2)

    def test_get(self):
        config = zirconium.ApplicationConfig(True)
        config.load_from_dict({
            "one": 1
        })
        self.assertEqual(config["one"], 1)
        self.assertRaises(ValueError, config.__getitem__, 'two')
        self.assertIsNone(config.get("two"))
        self.assertEqual(config.get("three", default=3), 3)
        self.assertRaises(ValueError, config.get, "four", raise_error=True)

    def test_dict_load(self):
        config = zirconium.ApplicationConfig(True)
        config.load_from_dict({
            "one": {
                "two": "three"
            }
        })
        self.assertTrue("one" in config)
        self.assertTrue(("one" ,"two") in config)
        self.assertEqual(config["one", "two"], "three")

    def test_bytes_coerce(self):
        config = zirconium.ApplicationConfig(True)
        config.load_from_dict({
            "raw": 512,
            "raw_str": "512",
            "for_def": 512,
            "bits": "512bit",
            "bytes": "512b",
            "kib": "2kib",
            "mib": "2mib",
            "gib": "2gib",
            "tib": "2tib",
            "pib": "2pib",
            "eib": "2eib",
            "k": "2K",
            "m": "2M",
            "g": "2G",
            "t": "2T",
            "p": "2P",
            "e": "2E",
            "kb": "2KB",
            "mb": "2MB",
            "gb": "2GB",
            "tb": "2TB",
            "pb": "2PB",
            "eb": "2EB",
            "no": "2WB",
            "very_no": "Foobar"
        })
        self.assertEqual(config.as_bytes("raw"), 512)
        self.assertEqual(config.as_bytes("for_def", default_units="kib"), 512 * 1024)
        self.assertEqual(config.as_bytes("raw_str"), 512)
        self.assertEqual(config.as_bytes("bits"), 64.0)
        self.assertEqual(config.as_bytes("bytes"), 512)
        self.assertEqual(config.as_bytes("kib"), 1024 * 2)
        self.assertEqual(config.as_bytes("mib"), 1024 * 1024 * 2)
        self.assertEqual(config.as_bytes("gib"), 1024 * 1024 * 1024 * 2)
        self.assertEqual(config.as_bytes("tib"), 1024 * 1024 * 1024 * 1024 * 2)
        self.assertEqual(config.as_bytes("pib"), 1024 * 1024 * 1024 * 1024 * 1024 * 2)
        self.assertEqual(config.as_bytes("eib"), 1024 * 1024 * 1024 * 1024 * 1024 * 1024 * 2)
        self.assertEqual(config.as_bytes("k"), 1024 * 2)
        self.assertEqual(config.as_bytes("m"), 1024 * 1024 * 2)
        self.assertEqual(config.as_bytes("g"), 1024 * 1024 * 1024 * 2)
        self.assertEqual(config.as_bytes("t"), 1024 * 1024 * 1024 * 1024 * 2)
        self.assertEqual(config.as_bytes("p"), 1024 * 1024 * 1024 * 1024 * 1024 * 2)
        self.assertEqual(config.as_bytes("e"), 1024 * 1024 * 1024 * 1024 * 1024 * 1024 * 2)
        self.assertEqual(config.as_bytes("kb"), 1024 * 2)
        self.assertEqual(config.as_bytes("mb"), 1024 * 1024 * 2)
        self.assertEqual(config.as_bytes("gb"), 1024 * 1024 * 1024 * 2)
        self.assertEqual(config.as_bytes("tb"), 1024 * 1024 * 1024 * 1024 * 2)
        self.assertEqual(config.as_bytes("pb"), 1024 * 1024 * 1024 * 1024 * 1024 * 2)
        self.assertEqual(config.as_bytes("eb"), 1024 * 1024 * 1024 * 1024 * 1024 * 1024 * 2)
        self.assertEqual(config.as_bytes("kb", allow_metric=True), 1000 * 2)
        self.assertEqual(config.as_bytes("mb", allow_metric=True), 1000 * 1000 * 2)
        self.assertEqual(config.as_bytes("gb", allow_metric=True), 1000 * 1000 * 1000 * 2)
        self.assertEqual(config.as_bytes("tb", allow_metric=True), 1000 * 1000 * 1000 * 1000 * 2)
        self.assertEqual(config.as_bytes("pb", allow_metric=True), 1000 * 1000 * 1000 * 1000 * 1000 * 2)
        self.assertEqual(config.as_bytes("eb", allow_metric=True), 1000 * 1000 * 1000 * 1000 * 1000 * 1000 * 2)

    def test_timedelta_coerce(self):
        config = zirconium.ApplicationConfig(True)
        config.load_from_dict({
            "def": 234,
            "def_min": 234,
            "test_s": "23s",
            "test_m": "23m",
            "test_h": "23h",
            "test_d": "23d",
            "test_w": "23w",
            "test_w1s": "23 w",
            "test_w3s": "   23     w  ",
            "test_us": "23us",
            "test_ms": "23ms",
            "test_str_no": "23",
            "test_err": "23n",
            "test_big_err": "foobar"
        })
        self.assertEqual(config.as_timedelta("def"), datetime.timedelta(seconds=234))
        self.assertEqual(config.as_timedelta("def_min", default_units="m"), datetime.timedelta(minutes=234))
        self.assertEqual(config.as_timedelta("test_s"), datetime.timedelta(seconds=23))
        self.assertEqual(config.as_timedelta("test_m"), datetime.timedelta(minutes=23))
        self.assertEqual(config.as_timedelta("test_h"), datetime.timedelta(hours=23))
        self.assertEqual(config.as_timedelta("test_d"), datetime.timedelta(days=23))
        self.assertEqual(config.as_timedelta("test_w"), datetime.timedelta(weeks=23))
        self.assertEqual(config.as_timedelta("test_w1s"), datetime.timedelta(weeks=23))
        self.assertEqual(config.as_timedelta("test_w3s"), datetime.timedelta(weeks=23))
        self.assertEqual(config.as_timedelta("test_us"), datetime.timedelta(microseconds=23))
        self.assertEqual(config.as_timedelta("test_ms"), datetime.timedelta(milliseconds=23))
        self.assertEqual(config.as_timedelta("test_str_no"), datetime.timedelta(seconds=23))
        self.assertRaises(ValueError, config.as_timedelta, "test_err")
        self.assertRaises(ValueError, config.as_timedelta, "test_big_err")


    def test_int_coerce(self):
        config = zirconium.ApplicationConfig(True)
        config.load_from_dict({
            "str": "1234",
            "int": 12,
            "bad": "modern general",
            "nope": None,
            "blank": "",
        })
        self.assertEqual(config.as_int("str"), 1234)
        self.assertEqual(config.as_int("int"), 12)
        self.assertIsNone(config.as_int("nope"))
        self.assertIsNone(config.as_int("blank"))
        self.assertRaises(ValueError, config.as_int, "bad")

    def test_int_ref_coerce(self):
        config = zirconium.ApplicationConfig(True)
        config.load_from_dict({
            "str": "1234",
            "str_int": 1234,
            "int": 12,
            "falsy": 0,
            "bad": "modern general",
            "nope": None,
            "blank": "",
        })
        self.assertEqual(config.as_int_ref("str"), config.as_int_ref("str_int"))
        self.assertNotEqual(config.as_int_ref("str"), config.as_int_ref("int"))
        self.assertEqual(config.as_int_ref("str"), 1234)
        self.assertEqual(config.as_int_ref("int"), 12)
        self.assertIsNone(config.as_int_ref("nope").raw_value())
        self.assertIsNone(config.as_int_ref("blank").raw_value())
        self.assertTrue(config.as_int_ref("nope").is_none())
        self.assertTrue(config.as_int_ref("blank").is_none())
        self.assertFalse(config.as_int_ref("str").is_none())
        self.assertRaises(ValueError, config.as_int_ref("bad").raw_value)
        self.assertTrue(config.as_int_ref("int"))
        self.assertFalse(config.as_int_ref("falsy"))

    def test_cache_by_type(self):
        config = zirconium.ApplicationConfig(True)
        config.load_from_dict({
            "int": 12,
        })
        self.assertEqual(config.as_str("int"), "12")
        self.assertEqual(config.as_int("int"), 12)

    def test_float_coerce(self):
        config = zirconium.ApplicationConfig(True)
        config.load_from_dict({
            "int": 123,
            "str_int": "1234",
            "str_float": "12.34",
            "float": 12.34,
            "exp": "1.1e3",
            "decimal": decimal.Decimal("12.34"),
            "bad": "modern general",
            "nope": None,
            "blank": "",
        })
        self.assertEqual(config.as_float("int"), 123.0)
        self.assertEqual(config.as_float("str_int"), 1234.0)
        self.assertEqual(config.as_float("str_float"), 12.34)
        self.assertEqual(config.as_float("float"), 12.34)
        self.assertEqual(config.as_float("exp"), 1100.0)
        self.assertEqual(config.as_float("decimal"), 12.34)
        self.assertIsNone(config.as_float("nope"))
        self.assertIsNone(config.as_float("blank"))
        self.assertRaises(ValueError, config.as_float, "bad")

    def test_decimal_coerce(self):
        config = zirconium.ApplicationConfig(True)
        config.load_from_dict({
            "int": 123,
            "str_int": "1234",
            "str_float": "12.34",
            "exp": "1.1e3",
            "decimal": decimal.Decimal("12.34"),
            "bad": "modern general",
            "nope": None,
            "blank": "",
        })
        self.assertEqual(config.as_decimal("int"), decimal.Decimal("123.0"))
        self.assertEqual(config.as_decimal("str_int"), decimal.Decimal("1234.0"))
        self.assertEqual(config.as_decimal("str_float"), decimal.Decimal("12.34"))
        self.assertEqual(config.as_decimal("decimal"), decimal.Decimal("12.34"))
        self.assertEqual(config.as_float("exp"), decimal.Decimal("1100"))
        self.assertIsNone(config.as_decimal("nope"))
        self.assertIsNone(config.as_decimal("blank"))
        self.assertRaises(decimal.InvalidOperation, config.as_decimal, "bad")

    def test_bool_coerce(self):
        config = zirconium.ApplicationConfig(True)
        config.load_from_dict({
            "zero": 0,
            "blank": "",
            "not_blank": "modern general",
            "true": True,
            "false": False,
            "one": 1,
            "positive": 512,
            "negative": -1,
            "nope": None,
            "empty_list": [],
            "empty_tuple": tuple(),
            "empty_set": set(),
            "empty_dict": dict(),
        })
        self.assertFalse(config.as_bool("zero"))
        self.assertFalse(config.as_bool("blank"))
        self.assertFalse(config.as_bool("false"))
        self.assertFalse(config.as_bool("nope"))
        self.assertFalse(config.as_bool("empty_list"))
        self.assertFalse(config.as_bool("empty_tuple"))
        self.assertFalse(config.as_bool("empty_set"))
        self.assertFalse(config.as_bool("empty_dict"))
        self.assertTrue(config.as_bool("not_blank"))
        self.assertTrue(config.as_bool("true"))
        self.assertTrue(config.as_bool("one"))
        self.assertTrue(config.as_bool("not_blank"))
        self.assertTrue(config.as_bool("positive"))
        self.assertTrue(config.as_bool("negative"))

    def test_date_coerce(self):
        config = zirconium.ApplicationConfig(True)
        est = datetime.timezone(-datetime.timedelta(hours=5), "EST")
        config.load_from_dict({
            "blank": "",
            "nope": None,
            "iso_str": "2020-01-01",
            "iso_dt_str": "2020-01-01 01:02:03",
            "date_obj": datetime.date(2020, 1, 1),
            "datetime_obj": datetime.datetime(2020, 1, 1, 0, 0, 0),
            "datetimetz_obj": datetime.datetime(2020, 1, 1, 0, 0, 0, tzinfo=est)
        })
        self.assertIsNone(config.as_date("blank"))
        self.assertIsNone(config.as_date("nope"))
        self.assertEqual(config.as_date("iso_str"), datetime.date(2020, 1, 1))
        self.assertEqual(config.as_date("iso_dt_str"), datetime.date(2020, 1, 1))
        self.assertEqual(config.as_date("date_obj"), datetime.date(2020, 1, 1))
        self.assertEqual(config.as_date("datetime_obj"), datetime.date(2020, 1, 1))
        self.assertEqual(config.as_date("datetimetz_obj"), datetime.date(2020, 1, 1))

    def test_datetime_coerce(self):
        config = zirconium.ApplicationConfig(True)
        est = datetime.timezone(-datetime.timedelta(hours=5), "EST")
        config.load_from_dict({
            "blank": "",
            "nope": None,
            "iso_str": "2020-01-01",
            "iso_dt_str": "2020-01-01 01:02:03",
            "date_obj": datetime.date(2020, 1, 1),
            "datetime_obj": datetime.datetime(2020, 1, 1, 1, 2, 3),
            "datetimetz_obj": datetime.datetime(2020, 1, 1, 1, 2, 3, tzinfo=est)
        })
        self.assertIsNone(config.as_datetime("blank"))
        self.assertIsNone(config.as_datetime("nope"))
        self.assertEqual(config.as_datetime("iso_str"), datetime.datetime(2020, 1, 1))
        self.assertEqual(
            config.as_datetime("iso_str", tzinfo=est),
            datetime.datetime(2020, 1, 1, tzinfo=est)
        )
        self.assertEqual(config.as_datetime("iso_dt_str"), datetime.datetime(2020, 1, 1, 1, 2, 3))
        self.assertEqual(
            config.as_datetime("iso_dt_str", tzinfo=est),
            datetime.datetime(2020, 1, 1, 1, 2, 3, tzinfo=est)
        )
        self.assertEqual(config.as_datetime("date_obj"), datetime.datetime(2020, 1, 1))
        self.assertEqual(
            config.as_datetime("date_obj", tzinfo=est),
            datetime.datetime(2020, 1, 1, tzinfo=est)
        )
        self.assertEqual(config.as_datetime("datetime_obj"), datetime.datetime(2020, 1, 1, 1, 2, 3))
        self.assertEqual(
            config.as_datetime("datetime_obj", tzinfo=est),
            datetime.datetime(2020, 1, 1, 1, 2, 3, tzinfo=est)
        )
        self.assertEqual(config.as_datetime("datetimetz_obj"), datetime.datetime(2020, 1, 1, 1, 2, 3, tzinfo=est))



import os
import decimal
import datetime
import threading
import sys
import time
from pathlib import Path
import zirconium.sproviders as sp
import logging
import typing as t


from autoinject import injector, CacheStrategy
from .parsers import JsonConfigParser, IniConfigParser, YamlConfigParser, TomlConfigParser, CfgConfigParser
from .utils import MutableDeepDict, _AppConfigHooks, convert_to_timedelta, convert_to_bytes, parse_for_units

# Metadata entrypoint support depends on Python version
import importlib.util
if importlib.util.find_spec("importlib.metadata"):
    # Python 3.10 supports entry_points(group=?)
    if sys.version_info.minor >= 10:
        from importlib.metadata import entry_points
    # Python 3.8 and 3.9 have metadata, but don't support the keyword argument
    else:
        from importlib.metadata import entry_points as _entry_points

        def entry_points(group=None):
            eps = _entry_points()
            if group is None:
                return eps
            elif group in eps:
                return eps[group]
            else:
                return []

# Backwards support for Python 3.7
else:
    from importlib_metadata import entry_points


VT = t.TypeVar("VT")


class _ConfigRef(t.Generic[VT]):

    def __init__(self, cfg_obj, cb_method, key, **kwargs):
        self.config = cfg_obj
        self.cb_method = cb_method
        self.key = key
        self.kwargs = kwargs
        self._cached = None
        self._cache_identifier = None

    def _ensure_cache(self) -> t.Optional[VT]:
        if self._cache_identifier is None or self._cache_identifier != self.config.cache_identifier:
            self._cache_identifier = self.config.cache_identifier
            self._cached = getattr(self.config, self.cb_method)(self.key, **self.kwargs)
        return self._cached

    def __eq__(self, other) -> bool:
        if isinstance(other, _ConfigRef):
            return self.raw_value() == other.raw_value()
        else:
            return self.raw_value() == other

    def raw_value(self) -> t.Optional[VT]:
        return self._ensure_cache()

    def is_none(self) -> bool:
        return self._ensure_cache() is None

    def __getattr__(self, item):
        c = self._ensure_cache()
        if c:
            return getattr(c, item)
        return None

    def __bool__(self) -> bool:
        return bool(self._ensure_cache())

    def __str__(self) -> str:
        return str(self._ensure_cache())

    def __int__(self) -> int:
        return int(self._ensure_cache())

    def __float__(self) -> float:
        return float(self._ensure_cache())


@injector.register("zirconium.config.ApplicationConfig", caching_strategy=CacheStrategy.GLOBAL_CACHE)
class ApplicationConfig(MutableDeepDict):

    ach: _AppConfigHooks = None

    @injector.construct
    def __init__(self, manual_init=False):
        super().__init__()
        self.log = logging.getLogger("zirconium")
        self.encoding = "utf-8"
        self.parsers = [
            TomlConfigParser(),
            YamlConfigParser(),
            CfgConfigParser(),
            IniConfigParser(),
            JsonConfigParser(),
        ]
        self._secret_providers = {}
        if sp.AZURE_ENABLED:
            self._secret_providers["azure_key_vault"] = sp.azure_key_vault
        self.file_registry = {
            "defaults": [],
            "regulars": [],
            "environment": [],
        }
        self.secrets_env_map = {}
        self.secrets_map = {}
        self.environment_map = {}
        self._default_config = {}
        self._on_load = []
        self.loaded_files = []
        self._init_flag = False
        self.cache_identifier = None
        self._cached_gets = {}
        self.registry_lock = threading.RLock()
        self.cache_lock = threading.RLock()
        if not manual_init:
            auto_register = entry_points(group="zirconium.parsers")
            for ep in auto_register:
                self.parsers.append(ep.load()(self))
            auto_register = entry_points(group="zirconium.config")
            for ep in auto_register:
                registrar_func = ep.load()
                registrar_func(self)
            self.ach.execute_hooks(self)
            self.init()

    def on_load(self, cb):
        self._on_load.append(cb)

    def get(self, *key, default=None, coerce=None, blank_to_none=False, raw=False, raise_error=False):
        key = self._expand_key(key)
        value = super().get(*key, default=default, raise_error=raise_error)
        if blank_to_none and value == "":
            value = None
        if (not raw) and isinstance(value, str):
            value = self.resolve_environment_references(value)
        if coerce and value is not None:
            value = coerce(value)
        return value

    def get_ref(self, *key, **kwargs) -> _ConfigRef[t.Any]:
        return _ConfigRef[t.Any](self, 'get', key, **kwargs)

    def resolve_environment_references(self, value: str) -> str:
        pos = 0
        # Escape rule for $ is dollar signs that precede a bracket
        # ${var}
        # $${not_a_var}, equivalent to string ${not_a_var}
        # $$${var}, equivalent to "${}".format(os.environ["var"])
        # $$$${not_a var}, equivalent to string $${not_a_var}
        #
        # ${{\}} equivalent to os.environ["{}"]
        # ${{\\\}} equivalent to os.environ["{\}"]
        state = "buffering"
        ref_buffer = ""
        resolved = ""
        for i in range(0, len(value)):
            c = value[i]
            if c == "$":
                if state == "buffering":
                    state = "opening_ref"
                    continue
                elif state == "opening_ref":
                    resolved += "$"
                    state = "buffering"
                    continue
            elif c == "{":
                if state == "opening_ref":
                    state = "in_ref"
                    continue
            elif c == "\\":
                if state == "in_ref":
                    state = "escaping_in_ref"
                    continue
                elif state == "escaping_in_ref":
                    ref_buffer += "\\"
                    state = "in_ref"
                    continue
            elif c == "}":
                if state == "escaping_in_ref":
                    ref_buffer += "}"
                    state = "in_ref"
                    continue
                elif state == "in_ref":
                    resolved += self.parse_env_reference(ref_buffer, "")
                    ref_buffer = ""
                    state = "buffering"
                    continue
            if state == "buffering":
                resolved += c
            elif state == "in_ref":
                ref_buffer += c
            elif state == "escaping_in_ref":
                ref_buffer += "\\" + c
                state = "in_ref"
        if ref_buffer:
            resolved += "${" + ref_buffer
        return resolved

    def parse_env_reference(self, name, default_val=None):
        if "=" in name:
            default_val = name[name.find("=") + 1:]
            name = name[0:name.find("=")]
        actual_val = self.get_env_var(name)
        return default_val if actual_val is None else actual_val

    def get_env_var(self, env_var_name):
        if env_var_name in os.environ:
            return os.environ[env_var_name]
        elif env_var_name.lower() in os.environ:
            return os.environ[env_var_name.lower()]
        elif env_var_name.upper() in os.environ:
            return os.environ[env_var_name.upper()]
        elif env_var_name in self.secrets_env_map:
            return self.get_secret(*self.secrets_env_map[env_var_name])
        return None

    def get_secret(self, secret_path, secret_provider):
        if secret_provider not in self._secret_providers:
            self.log.warning(f"Secret provider {secret_provider} not found")
            return None
        return self._secret_providers[secret_provider](secret_path)

    def as_bytes(self, key: t.Union[t.Iterable, t.AnyStr], default=None, default_units: str = "b", allow_metric: bool = False, raw: bool = False) -> t.Union[int, float]:
        val = self.get(key, default=default, blank_to_none=True, raw=raw)
        if val is None:
            return val
        elif isinstance(val, int) or isinstance(val, float):
            return convert_to_bytes(val, default_units, not allow_metric)
        else:
            val, units = parse_for_units(str(val), 3, default_units)
            return convert_to_bytes(val, units, not allow_metric)

    def as_bytes_ref(self, key: t.Union[t.Iterable, t.AnyStr], default=None, default_units="s", raw=False) -> _ConfigRef[t.Union[int, float]]:
        return _ConfigRef[t.Union[int, float]](self, 'as_bytes', key, default=default, default_units=default_units, raw=raw)

    def as_timedelta(self, key: t.Union[t.Iterable, t.AnyStr], default=None, default_units: str = "s", raw: bool = False) -> t.Optional[datetime.timedelta]:
        val = self.get(key, default=default, blank_to_none=True, raw=raw)
        if val is None or isinstance(val, datetime.timedelta):
            return val
        elif isinstance(val, int) or isinstance(val, float):
            return convert_to_timedelta(val, default_units)
        else:
            val, units = parse_for_units(str(val), 2, default_units)
            return convert_to_timedelta(val, units)

    def as_timedelta_ref(self, key: t.Union[t.Iterable, t.AnyStr], default=None, default_units="s", raw=False) -> _ConfigRef[datetime.timedelta]:
        return _ConfigRef[datetime.timedelta](self, 'as_timedelta', key, default=default, default_units=default_units, raw=raw)

    def as_date(self, key: t.Union[t.Iterable, t.AnyStr], default=None, raw=False) -> t.Optional[datetime.date]:
        dt = self.get(key, default=default, blank_to_none=True, raw=raw)
        if isinstance(dt, datetime.datetime):
            return datetime.date(dt.year, dt.month, dt.day)
        elif dt is None or isinstance(dt, datetime.date):
            return dt
        else:
            try:
                return datetime.date.fromisoformat(dt)
            except ValueError:
                dt = datetime.datetime.fromisoformat(dt)
                return datetime.date(dt.year, dt.month, dt.day)

    def as_date_ref(self, key: t.Union[t.Iterable, t.AnyStr], default=None, raw=False) -> _ConfigRef[datetime.date]:
        return _ConfigRef[datetime.date](self, 'as_date', key, default=default, raw=raw)

    def as_datetime(self, key: t.Union[t.Iterable, t.AnyStr], default=None, tzinfo=None, raw=False) -> t.Optional[datetime.datetime]:
        dt = self.get(key, default=default, blank_to_none=True, raw=raw)
        if dt is None:
            return None
        if isinstance(dt, datetime.datetime):
            if dt.tzinfo is None and tzinfo is not None:
                return datetime.datetime(
                    dt.year,
                    dt.month,
                    dt.day,
                    dt.hour,
                    dt.minute,
                    dt.second,
                    dt.microsecond,
                    tzinfo
                )
            return dt
        elif isinstance(dt, datetime.date):
            return datetime.datetime(dt.year, dt.month, dt.day, 0, 0, 0, tzinfo=tzinfo)
        else:
            dt = datetime.datetime.fromisoformat(dt)
            if dt.tzinfo is None and tzinfo is not None:
                return datetime.datetime(
                    dt.year,
                    dt.month,
                    dt.day,
                    dt.hour,
                    dt.minute,
                    dt.second,
                    dt.microsecond,
                    tzinfo
                )
            return dt

    def as_datetime_ref(self, key: t.Union[t.Iterable, t.AnyStr], default=None, tzinfo=None, raw=False) -> _ConfigRef[datetime.datetime]:
        return _ConfigRef[datetime.datetime](self, 'as_datetime', key, default=default, tzinfo=tzinfo, raw=raw)

    def as_int(self, key: t.Union[t.Iterable, t.AnyStr], default=None, raw=False) -> t.Optional[int]:
        return self.get(key, default=default, coerce=int, blank_to_none=True, raw=raw)

    def as_int_ref(self, key: t.Union[t.Iterable, t.AnyStr], default=None, raw=False) -> _ConfigRef[int]:
        return _ConfigRef[int](self, 'as_int', key, default=default, raw=raw)

    def as_float(self, key: t.Union[t.Iterable, t.AnyStr], default=None, raw=False) -> t.Optional[float]:
        return self.get(key, default=default, coerce=float, blank_to_none=True, raw=raw)

    def as_float_ref(self, key: t.Union[t.Iterable, t.AnyStr], default=None, raw=False) -> _ConfigRef[float]:
        return _ConfigRef[float](self, 'as_float', key, default=default, raw=raw)

    def as_decimal(self, key: t.Union[t.Iterable, t.AnyStr], default=None, raw=False) -> t.Optional[decimal.Decimal]:
        return self.get(key, default=default, coerce=decimal.Decimal, blank_to_none=True, raw=raw)

    def as_decimal_ref(self, key: t.Union[t.Iterable, t.AnyStr], default=None, raw=False) -> _ConfigRef[decimal.Decimal]:
        return _ConfigRef[decimal.Decimal](self, 'as_decimal', key, default=default, raw=raw)

    def as_str(self, key: t.Union[t.Iterable, t.AnyStr], default=None, raw=False) -> t.Optional[str]:
        return self.get(key, default=default, coerce=str, raw=raw)

    def as_str_ref(self, key: t.Union[t.Iterable, t.AnyStr], default=None, raw=False) -> _ConfigRef[str]:
        return _ConfigRef[str](self, 'as_float', key, default=default, raw=raw)

    def as_bool(self, key: t.Union[t.Iterable, t.AnyStr], default=None, raw=False) -> t.Optional[bool]:
        return bool(self.get(key, default=default, raw=raw))

    def as_bool_ref(self, key: t.Union[t.Iterable, t.AnyStr], default=None, raw=False) -> _ConfigRef[bool]:
        return _ConfigRef[bool](self, 'as_bool', key, default=default, raw=raw)

    def as_path(self, key: t.Union[t.Iterable, t.AnyStr], default=None, raw=False) -> t.Optional[Path]:
        return self.get(key, default=default, coerce=Path, blank_to_none=True, raw=raw)

    def as_path_ref(self, key: t.Union[t.Iterable, t.AnyStr], default=None, raw=False) -> _ConfigRef[Path]:
        return _ConfigRef[Path](self, 'as_path', key, default=default, raw=raw)

    def as_set(self, key: t.Union[t.Iterable, t.AnyStr], default=None) -> t.Optional[set]:
        return self.get(key, default=default, coerce=set, blank_to_none=True, raw=True)

    def as_set_ref(self, key: t.Union[t.Iterable, t.AnyStr], default=None) -> _ConfigRef[set]:
        return _ConfigRef[set](self, 'as_set', key, default=default)

    def as_list(self, key: t.Union[t.Iterable, t.AnyStr], default=None) -> t.Optional[list]:
        return self.get(key, default=default, coerce=list, blank_to_none=True, raw=True)

    def as_list_ref(self, key: t.Union[t.Iterable, t.AnyStr], default=None) -> _ConfigRef[list]:
        return _ConfigRef[list](self, 'as_list', key, default=default)

    def as_dict(self, key: t.Union[t.Iterable, t.AnyStr], default=None) -> t.Optional[dict]:
        return self.get(key, default=default, coerce=MutableDeepDict, blank_to_none=True, raw=True)

    def as_dict_ref(self, key: t.Union[t.Iterable, t.AnyStr], default=None) -> _ConfigRef[dict]:
        return _ConfigRef[dict](self, 'as_dict', key, default=default)

    def set_default_encoding(self, enc):
        self.encoding = enc

    def is_truthy(self, key: t.Union[t.Iterable, t.AnyStr]) -> bool:
        if not self._init_flag:
            self.init()
        parent, k = self._navigate_to_item(key)
        return parent is not None and k in parent and bool(parent[k])

    def register_parser(self, parser):
        self.parsers.append(parser)

    def register_secret_provider(self, name, callback):
        self._secret_providers[name] = callback

    def register_files(self,
                       search_directories: t.Iterable[t.Union[Path, t.AnyStr]],
                       files: t.Optional[t.Iterable[str]] = None,
                       default_files: t.Optional[t.Iterable[str]] = None):
        for sd in search_directories:
            sd = sd if isinstance(sd, Path) else Path(sd)
            for file in files:
                self.register_file(sd / file)
            for default_file in default_files:
                self.register_default_file(sd / default_file)

    def register_default_file(self, file_path, weight=None, parser=None, encoding=None):
        with self.registry_lock:
            if weight is None:
                weight = self._next_weight("defaults")
            self.file_registry["defaults"].append((file_path, weight, parser, encoding))

    def register_file(self, file_path, weight=None, parser=None, encoding=None):
        with self.registry_lock:
            if weight is None:
                weight = self._next_weight("regulars")
            self.file_registry["regulars"].append((file_path, weight, parser, encoding))

    def register_file_from_environ(self, env_var_name, weight, parser=None, encoding=None):
        with self.registry_lock:
            if weight is None:
                weight = self._next_weight("environment")
            self.file_registry["environment"].append((env_var_name, weight, parser, encoding))

    def _next_weight(self, key):
        if self.file_registry[key]:
            return max(x[1] for x in self.file_registry[key]) + 1
        return 0

    def register_environ_var(self, env_var_name, *target_config):
        self.environment_map[env_var_name] = target_config

    def register_environ_map(self, env_map):
        self.environment_map.update(env_map)

    def reload_config(self):
        # We take all three locks to prevent any weird multi-threaded behaviour from happening. All writes are blocked until we are done the re-load except our own.
        with self.lock:
            with self.registry_lock:
                with self.cache_lock:
                    self._cached_gets = {}
                    self.loaded_files = []
                    self._init_flag = False
                    self.clear()
                    self.init()

    def register_secret_as_environ_var(self, secret_path, secret_provider, env_var_name):
        self.secrets_env_map[env_var_name] = (secret_path, secret_provider)

    def register_secret_config(self, secret_path, secret_provider, *target_config):
        secret_provider = secret_provider.lower()
        self.secrets_map[f"{secret_path}{secret_provider}"] = (secret_path, secret_provider, target_config)

    def init(self):
        with self.registry_lock:
            if not self._init_flag:
                new_conf = MutableDeepDict()
                new_conf.deep_update(self._default_config)
                self.file_registry["defaults"].sort(key=lambda x: x[1])
                for file, weight, parser, enc in self.file_registry["defaults"]:
                    self.load_file(new_conf, file, parser, enc)
                self.file_registry["regulars"].sort(key=lambda x: x[1])
                for file, weight, parser, enc in self.file_registry["regulars"]:
                    self.load_file(new_conf, file, parser, enc)
                self.file_registry["environment"].sort(key=lambda x: x[1])
                for env_name, weight, parser, enc in self.file_registry["environment"]:
                    env_val = self.get_env_var(env_name)
                    if env_val:
                        self.load_file(new_conf, env_val, parser, enc)
                for env_name, target_config in self.environment_map:
                    env_val = self.get_env_var(env_name)
                    if env_val is not None:
                        self.log.info(f"Setting config from environment variable {env_name}")
                        new_conf[target_config] = env_val
                    else:
                        self.log.debug(f"No environment variable set for {env_name}")
                for key in self.secrets_map:
                    spath, sprovider, target_config = self._secret_providers[key]
                    secret_val = self.get_secret(sprovider, spath)
                    if secret_val is not None:
                        self.log.info(f"Loading secret from {sprovider} {spath}")
                        new_conf[target_config] = secret_val
                self.d = new_conf.d
                self._init_flag = True
                if self.cache_identifier is None:
                    self.cache_identifier = 1
                else:
                    self.cache_identifier += 1
                for cb in self._on_load:
                    cb(self)

    def load_file(self, new_conf, file_path, parser=None, encoding=None):
        with self.registry_lock:
            if encoding is None:
                encoding = self.encoding
            file_path = Path(file_path).expanduser().absolute()
            if file_path in self.loaded_files:
                return
            if file_path.exists():
                if parser:
                    self.log.info(f"Loading config file {file_path}")
                    new_conf.deep_update(parser.read_dict(file_path, encoding))
                    self.loaded_files.append(file_path)
                else:
                    for parser in self.parsers:
                        if parser.handles(file_path.name):
                            self.log.info(f"Loading config file {file_path}")
                            new_conf.deep_update(parser.read_dict(file_path, encoding))
                            self.loaded_files.append(file_path)
                            break
                    else:
                        self.log.warning(f"No parser found for {file_path}")
            else:
                self.log.info(f"No config file found at {file_path}")

    def set_defaults(self, d):
        self._default_config.update(d)

    def load_from_dict(self, d):
        self.deep_update(d)


def test_with_config(key: t.Union[list, str, set, tuple, dict], value: t.Any = None):
    """Create a test fixture for ApplicationConfig (if one doesn't exist) and set a config value for it"""
    def _inner(fn):
        if not hasattr(fn, "_autoinject_fixtures"):
            fn._autoinject_fixtures = {}
        if ApplicationConfig not in fn._autoinject_fixtures:
            def _build_app_config():
                ac = ApplicationConfig(True)
                ac.set_defaults(fn._zirconium_test_config)
                ac.init()
                return ac
            fn._autoinject_fixtures[ApplicationConfig] = (None, _build_app_config)
        if not hasattr(fn, "_zirconium_test_config"):
            fn._zirconium_test_config = MutableDeepDict()
        if isinstance(key, dict):
            fn._zirconium_test_config.update(key)
        else:
            fn._zirconium_test_config[key] = value

        return fn

    return _inner



from .config import ApplicationConfig, test_with_config
from .parsers import JsonConfigParser, IniConfigParser, YamlConfigParser, TomlConfigParser, CfgConfigParser, DbConfigParser
from .utils import _config_decorator as configure
from .utils import print_config, convert_to_bytes, convert_to_timedelta

__version__ = "1.2.3"
